/**
 * This software is licensed under the Apache 2 license, quoted below.
 *
 * Copyright 2011 Julien Eluard
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     [http://www.apache.org/licenses/LICENSE-2.0]
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.github.jeluard.net;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.Proxy;
import java.net.URL;
import java.net.URLStreamHandler;
import java.net.URLStreamHandlerFactory;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

/**
 * Workarounds for {@link URLStreamHandler}.
 */
public final class URLStreamHandlers {

    static class PatchedURLStreamFactory implements URLStreamHandlerFactory {

        private final URLStreamHandlerFactory originalStreamHandlerFactory;

        public PatchedURLStreamFactory(final URLStreamHandlerFactory originalStreamHandlerFactory) {
            this.originalStreamHandlerFactory = originalStreamHandlerFactory;
        }

        @Override
        public URLStreamHandler createURLStreamHandler(final String protocol) {
            final URLStreamHandler originalStreamHandler = this.originalStreamHandlerFactory.createURLStreamHandler(protocol);
            if (originalStreamHandler == null) {
                //Original factory returned null for this protocol, propagate it.
                if (URLStreamHandlers.LOGGER.isLoggable(Level.FINE)) {
                    URLStreamHandlers.LOGGER.log(Level.FINE, "Intercepting <{0}>", protocol);
                }

                return null;
            }

            final Enhancer enhancer = new Enhancer();
            enhancer.setSuperclass(originalStreamHandler.getClass());
            enhancer.setCallback(new MethodInterceptor() {
                @Override
                public Object intercept(final Object object, final Method method, final Object[] arguments, final MethodProxy proxy) throws Throwable {
                    if (URLStreamHandlers.LOGGER.isLoggable(Level.FINE)) {
                        URLStreamHandlers.LOGGER.log(Level.FINE, "Intercepting <{0}>", method);
                    }

                    //If openConnection(URL, Proxy) is called we first try to call it regularly.
                    //If this fails we fail back to openConnection(URL).
                    if (URLStreamHandlers.isOpenConnectionProxyMethod(method)) {
                        try {
                            return proxy.invokeSuper(object, arguments);
                        } catch (UnsupportedOperationException e) {
                            if (URLStreamHandlers.LOGGER.isLoggable(Level.FINE)) {
                                URLStreamHandlers.LOGGER.log(Level.FINE, "Got Exception while calling openConnection(<"+arguments[0]+">, <"+arguments[1]+">)", e);
                            }

                            return URLStreamHandlers.OPEN_CONNECTION_METHOD.invoke(object, arguments[0]);
                        }
                    } else {
                        return proxy.invokeSuper(object, arguments);
                    }
                }
            });
            return (URLStreamHandler) enhancer.create();
        }
    }

    private static final Logger LOGGER = Logger.getLogger(URLStreamHandlers.class.getPackage().getName());

    private static final String URL_STREAM_HANDLER_FACTORY_FIELD_NAME = "factory";
    private static final String OPEN_CONNECTION_METHOD_NAME = "openConnection";
    private static final Method OPEN_CONNECTION_METHOD;
    private static final Method OPEN_CONNECTION_PROXY_METHOD;

    static {
        try {
            OPEN_CONNECTION_METHOD = URLStreamHandler.class.getDeclaredMethod(URLStreamHandlers.OPEN_CONNECTION_METHOD_NAME, URL.class);
            OPEN_CONNECTION_METHOD.setAccessible(true);
            OPEN_CONNECTION_PROXY_METHOD = URLStreamHandler.class.getDeclaredMethod(URLStreamHandlers.OPEN_CONNECTION_METHOD_NAME, URL.class, Proxy.class);
            OPEN_CONNECTION_PROXY_METHOD.setAccessible(true);
        } catch (Exception e) {
            throw new Error("Failed to initialize", e);
        }
    }

    private URLStreamHandlers() {
    }

    /**
     * Resets currently set {@link URLStreamHandlerFactory}. Necessary to be able to set a new one.
     * @throws Exception
     */
    public static void resetURLStreamHandlerFactory() throws Exception {
        final Field field = URL.class.getDeclaredField(URLStreamHandlers.URL_STREAM_HANDLER_FACTORY_FIELD_NAME);
        field.setAccessible(true);
        field.set(null, null);
    }

    /**
     * @return current {@link URLStreamHandlerFactory}
     * @throws Exception
     */
    public static URLStreamHandlerFactory getURLStreamHandlerFactory() throws Exception {
        final Field field = URL.class.getDeclaredField(URLStreamHandlers.URL_STREAM_HANDLER_FACTORY_FIELD_NAME);
        field.setAccessible(true);
        return (URLStreamHandlerFactory) field.get(null);
    }

    
    public static boolean isOpenConnectionProxyMethod(final Method method) {
        if (URLStreamHandlers.OPEN_CONNECTION_METHOD_NAME.equals(method.getName())) {
            final Class<?>[] parameterTypes = method.getParameterTypes();
            return parameterTypes.length == 2 && URL.class.equals(parameterTypes[0]) && Proxy.class.equals(parameterTypes[1]);
        }
        return false;
    }

    /**
     * Install a new {@link URLStreamHandlerFactory} that delegates to current one.
     * This factory makes sure that created {@link URLStreamHandler} have their {@link URLStreamHandler#openConnection(URL, Proxy)} hooked so that when not implemented
     * {@link URLStreamHandler#openConnection(URL)} is called.
     *
     * @throws Exception
     */
    public static void installOpenConnectionProxy() throws Exception {
        final URLStreamHandlerFactory originalStreamHandlerFactory = URLStreamHandlers.getURLStreamHandlerFactory();
        if (originalStreamHandlerFactory == null) {
            throw new IllegalArgumentException("Must be called after "+URLStreamHandlerFactory.class.getSimpleName()+" is set");
        }
        if (originalStreamHandlerFactory instanceof PatchedURLStreamFactory) {
            //Already installed, do nothing.
            return;
        }

        URLStreamHandlers.resetURLStreamHandlerFactory();

        //Install our patched URLStreamHandlerFactory that will create URLStreamHandler performing 'safe' openConnection calls.
        URL.setURLStreamHandlerFactory(new PatchedURLStreamFactory(originalStreamHandlerFactory));
    }

}